@@ -3,6 +3,7 @@
 *  ipod-device.c
 *  Copyright (C) 2005-2006 Novell, Inc. 
 *  Written by Aaron Bockover <aaron@abock.org>
+*             James Willcox <snorp@snorp.net>
 ****************************************************************************/
 
 /*  
@@ -35,168 +36,16 @@
 #include <ctype.h>
 #include <errno.h>
 #include <glib/gstdio.h>
+#include <libxml/tree.h>
+#include <libxml/parser.h>
+#include <libxml/xpath.h>
 
 #include <glibtop/fsusage.h>
 
 #include "hal-common.h"
 #include "ipod-device.h"
-
-#define GB 1024
-
-typedef struct _IpodModel {
-    const gchar *model_number;
-    const guint64 capacity;
-    guint model_type;
-    guint generation;
-} IpodModel;
-
-static const IpodModel ipod_model_table [] = {
-    /* Handle idiots who hose their iPod file system, or 
-       lucky people with iPods we don't yet know about*/
-    {"Invalid", 0,    MODEL_TYPE_INVALID,    UNKNOWN_GENERATION},
-    {"Unknown", 0,    MODEL_TYPE_UNKNOWN,    UNKNOWN_GENERATION},
-    
-    /* First Generation */
-    {"8513", 5  * GB, MODEL_TYPE_REGULAR,    FIRST_GENERATION},
-    {"8541", 5  * GB, MODEL_TYPE_REGULAR,    FIRST_GENERATION},
-    {"8697", 5  * GB, MODEL_TYPE_REGULAR,    FIRST_GENERATION},
-    {"8709", 10 * GB, MODEL_TYPE_REGULAR,    FIRST_GENERATION},
-    
-    /* Second Generation */
-    {"8737", 10 * GB, MODEL_TYPE_REGULAR,    SECOND_GENERATION},
-    {"8740", 10 * GB, MODEL_TYPE_REGULAR,    SECOND_GENERATION},
-    {"8738", 20 * GB, MODEL_TYPE_REGULAR,    SECOND_GENERATION},
-    {"8741", 20 * GB, MODEL_TYPE_REGULAR,    SECOND_GENERATION},
-    
-    /* Third Generation */
-    {"8976", 10 * GB, MODEL_TYPE_REGULAR,    THIRD_GENERATION},
-    {"8946", 15 * GB, MODEL_TYPE_REGULAR,    THIRD_GENERATION},
-    {"9460", 15 * GB, MODEL_TYPE_REGULAR,    THIRD_GENERATION},
-    {"9244", 20 * GB, MODEL_TYPE_REGULAR,    THIRD_GENERATION},
-    {"8948", 30 * GB, MODEL_TYPE_REGULAR,    THIRD_GENERATION},
-    {"9245", 40 * GB, MODEL_TYPE_REGULAR,    THIRD_GENERATION},
-    
-    
-    /* Fourth Generation */
-    {"9282", 20 * GB, MODEL_TYPE_REGULAR,    FOURTH_GENERATION},
-    {"9787", 25 * GB, MODEL_TYPE_REGULAR_U2, FOURTH_GENERATION},
-    {"9268", 40 * GB, MODEL_TYPE_REGULAR,    FOURTH_GENERATION},
-    {"A079", 20 * GB, MODEL_TYPE_COLOR,      FOURTH_GENERATION},
-    {"A127", 20 * GB, MODEL_TYPE_COLOR_U2,   FOURTH_GENERATION},
-    {"9830", 60 * GB, MODEL_TYPE_COLOR,      FOURTH_GENERATION},
-    
-    /* First Generation Mini */
-    {"9160", 4 * GB, MODEL_TYPE_MINI,        FIRST_GENERATION},
-    {"9436", 4 * GB, MODEL_TYPE_MINI_BLUE,   FIRST_GENERATION},
-    {"9435", 4 * GB, MODEL_TYPE_MINI_PINK,   FIRST_GENERATION},
-    {"9434", 4 * GB, MODEL_TYPE_MINI_GREEN,  FIRST_GENERATION},
-    {"9437", 4 * GB, MODEL_TYPE_MINI_GOLD,   FIRST_GENERATION},    
-
-    /* Second Generation Mini */
-    {"9800", 4 * GB, MODEL_TYPE_MINI,        SECOND_GENERATION},
-    {"9802", 4 * GB, MODEL_TYPE_MINI_BLUE,   SECOND_GENERATION},
-    {"9804", 4 * GB, MODEL_TYPE_MINI_PINK,   SECOND_GENERATION},
-    {"9806", 4 * GB, MODEL_TYPE_MINI_GREEN,  SECOND_GENERATION},
-    {"9801", 6 * GB, MODEL_TYPE_MINI,        SECOND_GENERATION},
-    {"9803", 6 * GB, MODEL_TYPE_MINI_BLUE,   SECOND_GENERATION},
-    {"9805", 6 * GB, MODEL_TYPE_MINI_PINK,   SECOND_GENERATION},
-    {"9807", 6 * GB, MODEL_TYPE_MINI_GREEN,  SECOND_GENERATION},    
-
-    /* Photo / Fourth Generation */
-    {"9829", 30 * GB, MODEL_TYPE_COLOR,      FOURTH_GENERATION},
-    {"9585", 40 * GB, MODEL_TYPE_COLOR,      FOURTH_GENERATION},
-    {"9586", 60 * GB, MODEL_TYPE_COLOR,      FOURTH_GENERATION},
-    {"9830", 60 * GB, MODEL_TYPE_COLOR,      FOURTH_GENERATION},
-    
-    /* Shuffle / Fourth Generation */
-    {"9724", GB / 2, MODEL_TYPE_SHUFFLE,     FOURTH_GENERATION},
-    {"9725", GB,     MODEL_TYPE_SHUFFLE,     FOURTH_GENERATION},
-    
-    /* Nano / Fifth Generation */
-    {"A004", 2 * GB, MODEL_TYPE_NANO_WHITE,  FIFTH_GENERATION},
-    {"A099", 2 * GB, MODEL_TYPE_NANO_BLACK,  FIFTH_GENERATION},
-    {"A005", 4 * GB, MODEL_TYPE_NANO_WHITE,  FIFTH_GENERATION},
-    {"A107", 4 * GB, MODEL_TYPE_NANO_BLACK,  FIFTH_GENERATION},
-
-    /* Video / Fifth Generation */
-    {"A002", 30 * GB, MODEL_TYPE_VIDEO_WHITE, FIFTH_GENERATION},
-    {"A146", 30 * GB, MODEL_TYPE_VIDEO_BLACK, FIFTH_GENERATION},
-    {"A003", 60 * GB, MODEL_TYPE_VIDEO_WHITE, FIFTH_GENERATION},
-    {"A147", 60 * GB, MODEL_TYPE_VIDEO_BLACK, FIFTH_GENERATION},
-    
-    /* HP iPods, need contributions for this table */
-    {"E436", 40 * GB, MODEL_TYPE_REGULAR, FOURTH_GENERATION},
-    
-    {NULL, 0, 0, 0}
-};
-
-static const gchar *ipod_model_name_table [] = {
-    "Invalid",
-    "Unknown",
-    "Color",
-    "Color U2",
-    "Grayscale",
-    "Grayscale U2",
-    "Mini (Silver)",
-    "Mini (Blue)",
-    "Mini (Pink)",
-    "Mini (Green)",
-    "Mini (Gold)",
-    "Shuffle",
-    "Nano (White)",
-    "Nano (Black)",
-    "Video (White)",
-    "Video (Black)",
-    NULL
-};
-
-static const IpodArtworkFormat ipod_color_artwork_info[] = {
-    {IPOD_COVER_SMALL,        56,  56, 1017},
-    {IPOD_COVER_LARGE,       140, 140, 1016},
-    {IPOD_PHOTO_SMALL,        42,  30, 1009},
-    {IPOD_PHOTO_LARGE,       130,  88, 1015},
-    {IPOD_PHOTO_FULL_SCREEN, 220, 176, 1013},
-    {IPOD_PHOTO_TV_SCREEN,   720, 480, 1019},
-    {-1,                      -1,  -1,   -1}
-};
-
-static const IpodArtworkFormat ipod_nano_artwork_info[] = {
-    {IPOD_COVER_SMALL,        42,  42, 1031},
-    {IPOD_COVER_LARGE,       100, 100, 1027},
-    {IPOD_PHOTO_LARGE,        42,  37, 1032},
-    {IPOD_PHOTO_FULL_SCREEN, 176, 132, 1023},
-    {-1,                      -1,  -1,   -1}
-};
-
-static const IpodArtworkFormat ipod_video_artwork_info[] = {
-    {IPOD_COVER_SMALL,       100, 100, 1028},
-    {IPOD_COVER_LARGE,       200, 200, 1029},
-    {IPOD_PHOTO_SMALL,        50,  41, 1036},
-    {IPOD_PHOTO_LARGE,       130,  88, 1015},
-    {IPOD_PHOTO_FULL_SCREEN, 320, 240, 1024},
-    {IPOD_PHOTO_TV_SCREEN,   720, 480, 1019},
-    {-1,                      -1,  -1,   -1}
-};
-
-/* This will be indexed using a value from the MODEL_TYPE enum */
-static const IpodArtworkFormat *ipod_artwork_info_table[] = {
-    NULL,                    /* Invalid       */
-    NULL,                    /* Unknown       */
-    ipod_color_artwork_info, /* Color         */
-    ipod_color_artwork_info, /* Color U2      */
-    NULL,                    /* Grayscale     */ 
-    NULL,                    /* Grayscale U2  */
-    NULL,                    /* Mini (Silver) */
-    NULL,                    /* Mini (Blue)   */
-    NULL,                    /* Mini (Pink)   */
-    NULL,                    /* Mini (Green)  */
-    NULL,                    /* Mini (Gold)   */
-    NULL,                    /* Shuffle       */
-    ipod_nano_artwork_info,  /* Nano (White)  */
-    ipod_nano_artwork_info,  /* Nano (Black)  */
-    ipod_video_artwork_info, /* Video (White) */
-    ipod_video_artwork_info  /* Video (Black) */
-};
+#include "glib-compat.h"
+#include "ipod-model-table.h"
 
 #define g_free_if_not_null(o) \
     if(o != NULL) {           \
@@ -220,12 +69,22 @@ static gboolean ipod_device_reload(IpodDevice *device);
 static void ipod_device_construct_paths(IpodDevice *device);
 static gboolean ipod_device_info_load(IpodDevice *device);
 static guint ipod_device_detect_model(IpodDevice *device);
+static void ipod_device_parse_serial(IpodDevice *device);
 static gboolean ipod_device_detect_volume_info(IpodDevice *device);
 static LibHalContext *ipod_device_hal_initialize();
 static void ipod_device_detect_volume_used(IpodDevice *device);
 static gboolean ipod_device_read_sysinfo(IpodDevice *device);
 static gboolean ipod_device_detect_writeable(IpodDevice *device);
     
+typedef struct _IpodSerialNumber {
+    gchar *serial;
+    gchar production_factory[3];
+    gint production_year;
+    gint production_week;
+    gint production_number;
+    gchar model_code[4];
+} IpodSerialNumber;
+    
 struct IpodDevicePrivate {
     LibHalContext *hal_context;
     
@@ -243,14 +102,19 @@ struct IpodDevicePrivate {
     gchar *user_name;
     gchar *host_name;
     
+    /* SysInfo+Extended Fields */
+    gchar *model_number;
+    gchar *firmware_version;
+    IpodSerialNumber serial_number;
+
+    /* Artwork Formats */
+    IpodArtworkFormat *artwork_formats;
+    
     /* Volume Size/Usage */
     guint64 volume_size;
     guint64 volume_available;
     guint64 volume_used;
     
-    /* System Info */
-    GData *sys_info_table;
-    
     gchar *volume_uuid;
     gchar *volume_label;
     
@@ -344,8 +208,7 @@ ipod_device_get_property(GObject *object, guint prop_id,
             break;
         case PROP_DEVICE_MODEL_STRING:
             g_value_set_string(value,
-                ipod_model_name_table[ipod_model_table[
-                    device->priv->model_index].model_type]);
+                ipod_model_name_table[ipod_model_table[device->priv->model_index].model_type]);
             break;
         case PROP_DEVICE_GENERATION:
             g_value_set_uint(value, 
@@ -379,16 +242,13 @@ ipod_device_get_property(GObject *object, guint prop_id,
             g_value_set_boolean(value, device->priv->is_new);
             break;
         case PROP_SERIAL_NUMBER:
-            g_value_set_string(value, g_datalist_get_data(
-                &device->priv->sys_info_table, "pszSerialNumber"));
+            g_value_set_string(value, device->priv->serial_number.serial);
             break;
         case PROP_MODEL_NUMBER:
-            g_value_set_string(value, g_datalist_get_data(
-                &device->priv->sys_info_table, "ModelNumStr"));
+            g_value_set_string(value, device->priv->model_number);
             break;
         case PROP_FIRMWARE_VERSION: {
-            gchar *firmware = g_datalist_get_data(&device->priv->sys_info_table, "buildID");
-            gchar *firmware_formatted = ipod_version_string_from_hex_string(firmware);
+            gchar *firmware_formatted = ipod_version_string_from_hex_string(device->priv->firmware_version);
             g_value_set_string(value, firmware_formatted);
             if(firmware_formatted != NULL) {
                 g_free(firmware_formatted);
@@ -404,9 +264,19 @@ ipod_device_get_property(GObject *object, guint prop_id,
             g_value_set_boolean(value, device->priv->can_write);
             break;
         case PROP_ARTWORK_FORMAT:
-            g_value_set_pointer(value, 
-                (gpointer)ipod_artwork_info_table[ipod_model_table[
-                    device->priv->model_index].model_type]);
+            g_value_set_pointer(value, (gpointer) device->priv->artwork_formats);
+            break;
+        case PROP_MANUFACTURER_ID:
+            g_value_set_string(value, device->priv->serial_number.production_factory);
+            break;
+        case PROP_PRODUCTION_YEAR:
+            g_value_set_uint(value, device->priv->serial_number.production_year);
+            break;
+        case PROP_PRODUCTION_WEEK:
+            g_value_set_uint(value, device->priv->serial_number.production_week);
+            break;
+        case PROP_PRODUCTION_INDEX:
+            g_value_set_uint(value, device->priv->serial_number.production_number);
             break;
         default:
             G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
@@ -500,6 +370,10 @@ ipod_device_class_init(IpodDeviceClass *klass)
     GParamSpec *device_generation_param;
     GParamSpec *advertised_capacity_param;
     GParamSpec *artwork_format_param;
+    GParamSpec *manufacturer_id_param;
+    GParamSpec *production_year_param;
+    GParamSpec *production_week_param;
+    GParamSpec *production_index_param;
 
     GObjectClass *class = G_OBJECT_CLASS(klass);
 
@@ -596,7 +470,23 @@ ipod_device_class_init(IpodDeviceClass *klass)
 
     artwork_format_param = g_param_spec_pointer("artwork-formats",
         "Artwork Format", "Support Artwork Formats", G_PARAM_READABLE);
-
+    
+    manufacturer_id_param = g_param_spec_string("manufacturer-id", 
+        "Manufacturer ID", "ID of the plant that manufactured this iPod",
+        NULL, G_PARAM_READABLE);
+        
+    production_year_param = g_param_spec_uint("production-year", 
+        "Production Year", "Year this iPod was produced",
+        0, G_MAXUINT, 0, G_PARAM_READABLE);
+        
+    production_week_param = g_param_spec_uint("production-week", 
+        "Production Week", "Week number of this iPod in its production year",
+        0, G_MAXUINT, 0, G_PARAM_READABLE);
+        
+    production_index_param = g_param_spec_uint("production-index", 
+        "Production Index", "Production index of this iPod in its production week",
+        0, G_MAXUINT, 0, G_PARAM_READABLE);
+    
     class->set_property = ipod_device_set_property;
     class->get_property = ipod_device_get_property;    
     
@@ -669,7 +559,18 @@ ipod_device_class_init(IpodDeviceClass *klass)
     
     g_object_class_install_property(class, PROP_ARTWORK_FORMAT,
         artwork_format_param);
+    
+    g_object_class_install_property(class, PROP_MANUFACTURER_ID,
+        manufacturer_id_param);
         
+    g_object_class_install_property(class, PROP_PRODUCTION_YEAR,
+        production_year_param);
+        
+    g_object_class_install_property(class, PROP_PRODUCTION_WEEK,
+        production_week_param);
+        
+    g_object_class_install_property(class, PROP_PRODUCTION_INDEX,
+        production_index_param);
 }
 
 static void
@@ -687,7 +588,6 @@ ipod_device_init(IpodDevice *device)
     device->priv->user_name = NULL;
     device->priv->host_name = NULL;
     device->priv->adv_capacity = NULL;
-    device->priv->sys_info_table = NULL;
     device->priv->volume_uuid = NULL;
     device->priv->volume_label = NULL;
     
@@ -697,6 +597,8 @@ ipod_device_init(IpodDevice *device)
     
     device->priv->is_new = FALSE;
     device->priv->can_write = FALSE;
+    
+    memset(&device->priv->serial_number, 0, sizeof(IpodSerialNumber));
 }
 
 static void
@@ -715,12 +617,11 @@ ipod_device_finalize(GObject *object)
     g_free_if_not_null(device->priv->adv_capacity);
     g_free_if_not_null(device->priv->volume_uuid);
     g_free_if_not_null(device->priv->volume_label);
+    g_free_if_not_null(device->priv->model_number);
+    g_free_if_not_null(device->priv->serial_number.serial);
+    g_free_if_not_null(device->priv->firmware_version);
+    g_free_if_not_null(device->priv->artwork_formats);
 
-    if(device->priv->sys_info_table != NULL) {
-        g_datalist_clear(&device->priv->sys_info_table);
-        device->priv->sys_info_table = NULL;
-    }
-    
     if(device->priv->hal_context != NULL) {
         libhal_ctx_shutdown(device->priv->hal_context, NULL);
         libhal_ctx_free(device->priv->hal_context);
@@ -788,11 +689,12 @@ ipod_device_reload(IpodDevice *device)
     ipod_device_construct_paths(device);
     
     device->priv->is_new = !ipod_device_info_load(device);
-    
+
     ipod_device_detect_volume_used(device);
     ipod_device_detect_writeable(device);
     ipod_device_read_sysinfo(device);
     ipod_device_detect_model(device);
+    ipod_device_parse_serial(device);
     
     return device->priv->model_index != 0;
 }
@@ -869,30 +771,263 @@ ipod_device_write_device_info_string(gchar *str, FILE *fd)
     g_free(unistr);
 }
 
+static char *
+ipod_device_get_extended_sysinfo_path(IpodDevice *device)
+{
+    DBusMessage *dmesg, *reply;
+    DBusError error;
+    GPtrArray *options;
+    DBusConnection *dbus_connection;
+    char *path;
+    char *udi;
+
+    path = g_strdup_printf ("%sDevice/SysInfoExtended", device->priv->control_path);
+
+    if (g_file_test (path, G_FILE_TEST_EXISTS)) {
+        return path;
+    }
+    
+    udi = device->priv->hal_volume_id;
+    
+    dbus_error_init(&error);
+    dbus_connection = dbus_bus_get(DBUS_BUS_SYSTEM, &error);
+    
+    if(dbus_error_is_set(&error)) {
+        dbus_error_free(&error);
+        g_free (path);
+        return NULL;
+    }
+
+    
+    if(!(dmesg = dbus_message_new_method_call("org.freedesktop.Hal", udi,
+        "org.freedesktop.Hal.Device.Volume", "WriteIpodInfo"))) {
+        g_free (path);
+        return NULL;
+    }
+    
+    options = g_ptr_array_new();
+
+    if(!dbus_message_append_args(dmesg, 
+          DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &options->pdata, options->len,
+        DBUS_TYPE_INVALID)) {
+        dbus_message_unref(dmesg);
+        g_ptr_array_free(options, TRUE);
+        g_free (path);
+        return NULL;
+    }
+
+    g_ptr_array_free(options, TRUE);
+    
+    if(!(reply = dbus_connection_send_with_reply_and_block(
+        dbus_connection, dmesg, -1, &error))) {
+        dbus_message_unref(dmesg);
+        dbus_error_free(&error);
+        g_free (path);
+        return NULL;
+    }
+
+    dbus_message_unref(dmesg);
+    dbus_message_unref(reply);
+
+    if (!g_file_test (path, G_FILE_TEST_EXISTS)) {
+        g_free (path);
+        return NULL;
+    }
+    
+    return path;
+}
+
+static char *
+ipod_device_get_dict_value (xmlXPathContextPtr xpath, const char *path, const char *text)
+{
+    xmlXPathObjectPtr xpath_item;
+    char *pathstr;
+    char *val = NULL;
+
+    pathstr = g_strdup_printf ("%s/key[text()='%s']/following-sibling::*[1]", path, text);
+    
+    xpath_item = xmlXPathEvalExpression((xmlChar*)pathstr, xpath);
+    if (xpath_item->nodesetval && xpath_item->nodesetval->nodeTab[0] != NULL) {
+        val = (char *)xmlNodeGetContent (xpath_item->nodesetval->nodeTab[0]);
+    }
+
+    g_free (pathstr);
+    xmlXPathFreeObject (xpath_item);
+    return val;
+}
+
+static IpodPixelFormat
+ipod_device_convert_pixel_format(const char *format)
+{
+    if (strcmp (format, "32767579") == 0) {
+        return IPOD_PIXEL_FORMAT_IYUV;
+    } else if (strcmp (format, "4C353635") == 0) {
+        return IPOD_PIXEL_FORMAT_RGB565;
+    } else if (strcmp (format, "42353635") == 0) {
+    return IPOD_PIXEL_FORMAT_RGB565_BE;
+    }
+    
+    return IPOD_PIXEL_FORMAT_UNKNOWN;
+}
+
+static void
+ipod_device_parse_artwork_formats (GList **list, xmlXPathObjectPtr xpath_item, 
+    IpodArtworkUsage usage)
+{
+    int i;
+    
+    g_return_if_fail(xpath_item != NULL);
+    g_return_if_fail(xpath_item->nodesetval != NULL);
+    
+    for (i = 0; i < xpath_item->nodesetval->nodeNr; i++) {
+        IpodArtworkFormat *format = g_new0(typeof (IpodArtworkFormat), 1);
+        xmlNodePtr dict = xpath_item->nodesetval->nodeTab[i];
+        xmlNodePtr child;
+        char *key = NULL;
+
+        format->usage = usage;
+
+        for (child = dict->children; child; child = child->next) {
+            xmlChar *value;
+        
+            if (child->type != XML_ELEMENT_NODE) {
+                continue;
+            }
+            
+            value = xmlNodeGetContent(child);
+            
+            if (strcmp ((char *)child->name, "key") == 0) {
+                    g_free_if_not_null (key);
+                key = g_strdup ((char *)value);
+            } else if (strcmp(key, "FormatId") == 0) {
+                format->correlation_id = atoi((char *)value);
+            } else if (strcmp(key, "RenderWidth") == 0) {
+                format->width = atoi((char *)value);
+            } else if (strcmp(key, "RenderHeight") == 0) {
+                format->height = atoi((char *)value);
+            } else if (strcmp(key, "PixelFormat") == 0) {
+                format->pixel_format = ipod_device_convert_pixel_format((char *)value);
+            } else if (strcmp(key, "Rotation") == 0) {
+                format->rotation = atoi((char *)value);
+            }
+
+            xmlFree (value);
+        }
+
+        /* FIXME: is this really always the case? */
+        format->image_size = format->width * format->height * 2;
+
+        g_free_if_not_null (key);
+
+        *list = g_list_prepend (*list, format);
+    }
+}
+
+static void
+ipod_device_read_artwork_info (IpodDevice *device, xmlXPathContextPtr xpath)
+{
+    GList *list = NULL;
+    GList *l;
+    xmlXPathObjectPtr xpath_item;
+    int i;
+    IpodArtworkFormat last_format = { -1, -1, -1, -1, -1, -1, -1 };
+
+    xpath_item = xmlXPathEvalExpression((xmlChar *)"/plist/dict/key[text()='ImageSpecifications']/following-sibling::*[1]//dict", xpath);
+    if(xpath_item != NULL) {
+        if(xpath_item->nodesetval != NULL) {
+            ipod_device_parse_artwork_formats (&list, xpath_item, IPOD_ARTWORK_PHOTO);
+        }
+        xmlXPathFreeObject (xpath_item);
+    }
+    
+    xpath_item = xmlXPathEvalExpression((xmlChar *)"/plist/dict/key[text()='AlbumArt']/following-sibling::*[1]//dict", xpath);
+    if(xpath_item != NULL) {
+        if(xpath_item->nodesetval != NULL) {
+            ipod_device_parse_artwork_formats (&list, xpath_item, IPOD_ARTWORK_COVER);
+        }
+        xmlXPathFreeObject (xpath_item);
+    }
+    
+    if (device->priv->artwork_formats != NULL) {
+        g_free (device->priv->artwork_formats);
+    }
+    
+    if(list != NULL && g_list_length(list) > 0) {
+        device->priv->artwork_formats = g_new0(IpodArtworkFormat, g_list_length (list) + 1);
+
+        i = 0;
+        for (l = list; l; l = l->next) {
+            device->priv->artwork_formats[i++] = *(IpodArtworkFormat *)l->data;
+        }
+
+        device->priv->artwork_formats[i] = last_format;
+    } else {
+        device->priv->artwork_formats = NULL;
+    }
+}
+
+static gboolean
+ipod_device_read_sysinfo_hal(IpodDevice *device)
+{
+    char *path;
+    xmlDocPtr doc;
+    xmlNodePtr root;
+    xmlXPathContextPtr xpath;
+
+    device->priv->serial_number.serial = libhal_device_get_property_string(device->priv->hal_context,
+									   device->priv->hal_volume_id,
+									   "info.ipod.serial_number", NULL);
+    
+    path = ipod_device_get_extended_sysinfo_path (device);
+    if (path == NULL) {
+        return FALSE;
+    }
+
+    doc = xmlParseFile (path);
+    g_free (path);
+    
+    if (doc == NULL) {
+        return FALSE;
+    }
+    
+    root = xmlDocGetRootElement(doc);
+    if(root == NULL) {
+        xmlFreeDoc(doc);
+        return FALSE;
+    }
+
+    xpath = xmlXPathNewContext(doc);
+
+    device->priv->serial_number.serial = ipod_device_get_dict_value (xpath, "/plist/dict", "SerialNumber");
+
+    ipod_device_read_artwork_info (device, xpath);
+
+    xmlXPathFreeContext (xpath);
+    xmlFreeDoc (doc);
+
+    return TRUE;
+}
+
 static gboolean
 ipod_device_read_sysinfo(IpodDevice *device)
 {
     gchar *path, buf[512];
     gboolean corrupt_check = FALSE;
+    gboolean hal_result;
     FILE *fd;
     
     g_return_val_if_fail(IS_IPOD_DEVICE(device), FALSE);
-    
+
+    hal_result = ipod_device_read_sysinfo_hal (device);
+
     path = g_strdup_printf("%sDevice/SysInfo", device->priv->control_path);
     
     fd = fopen(path, "r");
     if(fd == NULL) {
         g_free(path);
-        return FALSE;
+        return hal_result;
     }
     
-    if(device->priv->sys_info_table != NULL) {
-        g_datalist_clear(&device->priv->sys_info_table);
-        device->priv->sys_info_table = NULL;
-    }
-
-    g_datalist_init(&device->priv->sys_info_table);
-
     while(fgets(buf, sizeof(buf), fd)) {
         gint key_length, value_length;
         gchar *key, *value, *p;
@@ -939,10 +1074,16 @@ ipod_device_read_sysinfo(IpodDevice *device)
         if(value[0] == '0' && value[1] == 'x' && (p = strchr(value, ' ')) != NULL) {
             *p = '\0';
         }
-        
-        g_datalist_set_data_full(&device->priv->sys_info_table, 
-            key, g_strdup(value), g_free);
-        g_free(key);
+
+        if (strcmp (key, "pszSerialNumber") == 0) {
+            device->priv->serial_number.serial = g_strdup (value);
+        } else if (strcmp (key, "ModelNumStr") == 0) {
+            device->priv->model_number = g_strdup (value);
+        } else if (strcmp (key, "buildID") == 0) {
+            device->priv->firmware_version = g_strdup (value);
+        }
+
+        g_free (key);
     }
     
     fclose(fd);
@@ -995,14 +1136,19 @@ ipod_device_detect_writeable(IpodDevice *device)
     g_return_val_if_fail(IS_IPOD_DEVICE(device), FALSE);
     
     device->priv->can_write = FALSE;
+    
+    if(!libhal_device_property_exists(device->priv->hal_context, device->priv->hal_volume_id, 
+        "volume.is_mounted_read_only", NULL)) {
+        device->priv->can_write = !libhal_device_get_property_bool(device->priv->hal_context, 
+            device->priv->hal_volume_id, "volume.is_mounted_read_only", NULL);
+        return device->priv->can_write;
+    }
 
     itunes_dir = g_strdup_printf("%siTunes", device->priv->control_path);
     
-    if(!g_file_test(itunes_dir, G_FILE_TEST_IS_DIR)) {
-        if(g_mkdir_with_parents(itunes_dir, 0755) != 0) {
-            g_free(itunes_dir);
-            return FALSE;
-        }
+    if(g_mkdir_with_parents(itunes_dir, 0755) == -1) {
+        g_free(itunes_dir);
+        return FALSE;
     }
     
     itunesdb_path = g_strdup_printf("%s/iTunesDB", itunes_dir);
@@ -1044,7 +1190,7 @@ ipod_device_get_model_index_from_table(const gchar *_model_num)
     gchar *p = model_num;
     
     if(isalpha(model_num[0])) {
-        *p++;
+        p++;
     }
     
     for(i = 2; ipod_model_table[i].model_number != NULL; i++) {
@@ -1065,35 +1211,44 @@ ipod_device_detect_model(IpodDevice *device)
     guint64 adv, act;
     gint cap;
     const gchar *model_number;
-    const gchar *board_rev;
+    const gchar *serial_number_suffix;
+    const gchar *board_rev = NULL;
     
     g_return_val_if_fail(IS_IPOD_DEVICE(device), 0);
 
-    device->priv->model_index = 0;
-    model_number = g_datalist_get_data(&device->priv->sys_info_table, "ModelNumStr");
-    board_rev = g_datalist_get_data(&device->priv->sys_info_table, "boardHwSwInterfaceRev");
-
-    if(model_number == NULL && board_rev == NULL) {
-        /* Will be an iPod Shuffle; it doesn't have a SysInfo file */
+    /* We know we have an iPod ... we just don't know what it is */
+       device->priv->model_index = 1;
+    
+    model_number = device->priv->model_number;
+    serial_number_suffix = NULL;
+    
+    if (device->priv->serial_number.serial != NULL) {
+        serial_number_suffix = &device->priv->serial_number.serial[strlen(device->priv->serial_number.serial) - 3];
+    }
+    
+    if(model_number == NULL && serial_number_suffix != NULL) {
+        guint64 current_size_diff = 0;
+        
+        /* probably have the 'extended' sysinfo which doesn't have model */
+        /* we'll have to figure it out ourselves */
+        
         for(i = 2; ipod_model_table[i].model_number != NULL; i++) {
-            if(ipod_model_table[i].model_type != MODEL_TYPE_SHUFFLE) {
+            if(ipod_model_table[i].serial_numbers == NULL ||  
+                strstr(ipod_model_table[i].serial_numbers, serial_number_suffix) == NULL) {
                 continue;
             }
-            
-            cap = ipod_model_table[i].capacity;    
-            adv = cap * 1048576;
+
+            cap = ipod_model_table[i].capacity;
+            adv = cap * 1048576; /* convert to bytes */
             act = device->priv->volume_size;
 
-            if((adv - act) / 1048576 < 50) {
+            if (current_size_diff == 0 || abs(adv - act) < current_size_diff) {
+                current_size_diff = abs(adv - act);
                 device->priv->model_index = i;
-                
-                g_datalist_set_data_full(&device->priv->sys_info_table, "ModelNumStr", 
-                    g_strdup_printf("M%s", ipod_model_table[i].model_number), g_free);
-                    
+                device->priv->model_number = g_strdup (ipod_model_table[i].model_number);
                 device->priv->adv_capacity = CAPACITY_DISPLAY_STRING(cap);
-                break;
             }
-        }    
+        }
     } else if(model_number == NULL) {
         /* Guess the model number based on capacity and generation;
            2G iPods and older do not have ModelNumStr in SysInfo */
@@ -1112,8 +1267,7 @@ ipod_device_detect_model(IpodDevice *device)
         }
         
         if(model_num != NULL) {
-            g_datalist_set_data_full(&device->priv->sys_info_table, "ModelNumStr", 
-                g_strdup_printf("M%s", model_num), g_free);
+            device->priv->model_number = g_strdup_printf("M%s", model_num);
             
             for(i = 2; ipod_model_table[i].model_number != NULL; i++) {
                 if(strcmp(ipod_model_table[i].model_number, model_num) == 0) {
@@ -1130,10 +1284,46 @@ ipod_device_detect_model(IpodDevice *device)
         cap = ipod_model_table[device->priv->model_index].capacity;
         device->priv->adv_capacity = CAPACITY_DISPLAY_STRING(cap);
     }
-        
+
     return device->priv->model_index;
 }
 
+static void
+ipod_device_parse_serial(IpodDevice *device)
+{
+    gchar *serial;
+    gchar tmp_serial[9];
+    IpodSerialNumber *decoded;
+    
+    g_return_if_fail(IS_IPOD_DEVICE(device));
+    
+    decoded = &device->priv->serial_number;
+    g_return_if_fail(decoded != NULL);
+    
+    serial = decoded->serial;
+    if(serial == NULL) {
+        return;
+    }
+    
+    g_return_if_fail(strlen(serial) == 11);
+    
+    memcpy(tmp_serial, serial, 8);
+    memcpy(decoded->production_factory, serial, 2);
+    
+    tmp_serial[8] = 0;
+    decoded->production_factory[2] = 0;
+    
+    decoded->production_year = (serial[2] - '0') + 2000;
+    decoded->production_week = (serial[3] - '0') * 10 + (serial[4] - '0');
+    
+    memcpy(decoded->model_code, serial + 8, 3);
+    decoded->model_code[3] = 0;
+    
+    decoded->production_number = strtol(tmp_serial + 5, NULL, 34);
+    
+    return;
+}
+
 static gboolean
 ipod_device_detect_volume_info(IpodDevice *device)
 {
@@ -1238,7 +1428,7 @@ ipod_device_detect_volume_info(IpodDevice *device)
             hal_context, device->priv->hal_volume_id, 
             "volume.mount_point", NULL);
     }
-    
+
     if(libhal_device_property_exists(hal_context, device->priv->hal_volume_id,
         "volume.size", NULL)) {
         device->priv->volume_size = libhal_device_get_property_uint64(
@@ -1272,6 +1462,7 @@ ipod_device_detect_volume_used(IpodDevice *device)
 {
     glibtop_fsusage fsbuf;
     glibtop_get_fsusage(&fsbuf, device->priv->mount_point);
+
     device->priv->volume_available = fsbuf.bfree * fsbuf.block_size;
     device->priv->volume_used = device->priv->volume_size - 
         device->priv->volume_available;
@@ -1288,8 +1479,8 @@ ipod_device_eject_hal(IpodDevice *device, gchar **error_out)
     DBusError error;
     GPtrArray *options;
     DBusConnection *dbus_connection;
-	
-	gchar *udi = device->priv->hal_volume_id;
+    
+    gchar *udi = device->priv->hal_volume_id;
 
     *error_out = NULL;
 
@@ -1673,20 +1864,18 @@ ipod_device_save(IpodDevice *device, GError **error_out)
     itunes_dir = g_strdup_printf("%siTunes", device->priv->control_path);
     path = g_strdup_printf("%s/DeviceInfo", itunes_dir);
     
-    if(!g_file_test(itunes_dir, G_FILE_TEST_IS_DIR)) {
-        if(g_mkdir_with_parents(itunes_dir, 0744) != 0) {
-            if(error_out != NULL) {
-                error = g_error_new(g_quark_from_static_string("IPOD_DEVICE"),
-                    ERROR_SAVE, "Could not create iTunes Directory: %s", 
-                    itunes_dir);
-                g_propagate_error(error_out, error);
-            }
-            
-            g_free(path);
-            g_free(itunes_dir);
-            
-            return FALSE;
+    if(g_mkdir_with_parents(itunes_dir, 0744) == -1) {
+        if(error_out != NULL) {
+            error = g_error_new(g_quark_from_static_string("IPOD_DEVICE"),
+                ERROR_SAVE, "Could not create iTunes Directory: %s", 
+                itunes_dir);
+            g_propagate_error(error_out, error);
         }
+        
+        g_free(path);
+        g_free(itunes_dir);
+        
+        return FALSE;
     }
     
     fd = fopen(path, "w+");
@@ -1737,17 +1926,13 @@ ipod_device_debug(IpodDevice *device)
     };
     
     static const gchar *artwork_names [] = {
-        "Small Cover:           ",
-        "Large Cover:           ",
-        "Small Photo:           ",
-        "Large Photo:           ",
-        "Full iPod Screen Photo:",
-        "Full TV Screen Photo:  "
+        "Photo:           ",
+        "Cover:           "
     };
 
     gchar *device_path, *mount_point, *control_path, *hal_id;
-    gchar *model_number, *adv_capacity, *model_string;
-    guint model, generation;
+    gchar *model_number, *adv_capacity, *model_string, *production_factory;
+    guint model, generation, production_year, production_week, production_index;
     gboolean is_new, can_write;
     gchar *serial_number, *firmware_version;
     guint64 volume_size, volume_used, volume_available;
@@ -1780,6 +1965,10 @@ ipod_device_debug(IpodDevice *device)
         "user-name", &user_name,
         "host-name", &host_name,
         "artwork-formats", &artwork_formats,
+        "manufacturer-id", &production_factory,
+        "production-year", &production_year,
+        "production-week", &production_week,
+        "production-index", &production_index,
         NULL);
         
     g_printf("Path Info\n");
@@ -1794,9 +1983,13 @@ ipod_device_debug(IpodDevice *device)
     g_printf("   iPod Generation:  %s\n", generation_names[generation]);
     g_printf("   Adv. Capacity:    %s\n", adv_capacity);
     g_printf("   Is New:           %s\n", is_new ? "YES" : "NO");
-    g_printf("   Writeable:        %s\n", can_write ? "YES" : "NO");
+    g_printf("   Writable:         %s\n", can_write ? "YES" : "NO");
     g_printf("   Serial Number:    %s\n", serial_number);
     g_printf("   Firmware Version: %s\n", firmware_version);
+    g_printf("   Manufacturer ID:  %s\n", production_factory);
+    g_printf("   Production Year:  %04d\n", production_year);
+    g_printf("   Production Week:  %02d\n", production_week);
+    g_printf("   Production Index: %d\n", production_index);
     
     g_printf("Volume Info\n");
     g_printf("   Volume Size:      %"G_GUINT64_FORMAT"\n", volume_size);
@@ -1815,8 +2008,8 @@ ipod_device_debug(IpodDevice *device)
 
         g_print("Supported Artwork Formats\n");
         format = artwork_formats;
-        while(format->type != -1) {
-            printf("   %s %dx%d\n", artwork_names[format->type],
+        while(format->usage != -1) {
+            printf("   %s %dx%d\n", artwork_names[format->usage],
                 format->width, format->height);
             format++;
         }
@@ -1825,3 +2018,23 @@ ipod_device_debug(IpodDevice *device)
     g_printf("\n");
     fflush(stdout);
 }
+
+void ipod_device_dump_model_table_as_xml()
+{
+    gint i;
+    
+    g_printf("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
+    g_printf("<ipod-models>\n");
+    for(i = 2; ipod_model_table[i].model_number != NULL; i++) {
+        g_printf("<model>\n");
+        g_printf("  <number>%s</number>\n", ipod_model_table[i].model_number); 
+        g_printf("  <name>%s</name>\n", ipod_model_name_table[ipod_model_table[i].model_type]);
+        g_printf("  <capacity>%"G_GUINT64_FORMAT"</capacity>\n", ipod_model_table[i].capacity);
+        g_printf("  <generation>%d</generation>\n", ipod_model_table[i].generation);
+        if(ipod_model_table[i].serial_numbers != NULL) {
+            g_printf("  <serials>%s</serials>\n", ipod_model_table[i].serial_numbers);
+        }
+        g_printf("</model>\n");
+    }
+    g_printf("</ipod-models>\n");
+}
